//	WLA - Wavefile Length Adjust
//
//	Copyright (C) 2001 Joakim Hjort
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


#include "stdafx.h"

const char Version[] = "0.26";

// Buffersize when reading and writing wavefiles
#define FILEIOBUFFERSIZE (100*1024)
#define FOURCC_WAVE mmioFOURCC('W', 'A', 'V', 'E')
#define FOURCC_data mmioFOURCC('d', 'a', 't', 'a')
#define FOURCC_fmt  mmioFOURCC('f', 'm', 't', ' ')


HMMIO hWI;								// Input multimedia file handle
MMCKINFO WIGH;						// For Group Header
MMCKINFO WISC;						// For finding chunks within the Group
WAVEFORMATEX WIFormat;		// For reading a fmt chunk's data
long WISamples = 0;				// # of samples in input wave
long WIMax = 0;						// Input amplityde max value

HMMIO hWO;								// Output multimedia file handle
MMCKINFO WOGH;						// For Group Header
MMCKINFO WOSC;						// For writing chunks within the Group
long WOSamples = 0;				// # of samples in output wave
long WOSilence = 0;				// # of silence samples added to output file
long WODrop = 0;					// # of samples to drop from output file

int Ration1 = 0;					// Ration of output file or new length
int Ration2 = 0;					// Ration of input file
BOOL fNormalize = FALSE;	// Normalize amplitude of input wavefile
BOOL fQuite = FALSE;			// Be quiet


///////////////////////////////////////////////////
// Normalize input wave to maximum amplification
// Returns TRUE on success.
BOOL Normalize(char *szInputFileName) {
	long Sample;      // One audio sample (both chanels)
	long L, R;
	long i, n, j = 0, p = 0;

	// Open input wavefile
	if (!(hWI = mmioOpen(szInputFileName, NULL, MMIO_READ | MMIO_DENYWRITE)))	{
		printf("Error opening input wavefile '%s'.\n\n", szInputFileName);
		return FALSE;
	}

	// Set a large read buffer
	if (mmioSetBuffer(hWI, NULL, FILEIOBUFFERSIZE, 0)) {
		printf("Error allocating write buffer!\n\n");
		return FALSE;
	}

	// Locate a WAVE Group header in the file, and read it in.
	// This marks the start of any embedded WAVE format within the file.
	WIGH.fccType = FOURCC_WAVE; 
	if (mmioDescend(hWI, (LPMMCKINFO)&WIGH, NULL, MMIO_FINDRIFF)) {
		printf("Error reading input wavefile - is it a wavefile?\n\n");
		return FALSE;
	}

	// Locate "data" chunk it contains the audio information
	WISC.ckid = FOURCC_data; 
	if (mmioDescend(hWI, &WISC, &WIGH, MMIO_FINDCHUNK)) { 
		printf("Error - input wavefile has no data chunk!"); 
		return FALSE;
	}

	// Normalize
	if (!fQuite) {
		ldiv_t t;
		
		t = ldiv(WISamples, 100);
		n = t.quot;
		printf(" reading input for normalization... 0%%\r");
	}

	for (i = 0; i < WISamples; i++) {
		// Read sample
		if (mmioRead(hWI, (HPSTR)&Sample, sizeof(Sample)) != (LRESULT)sizeof(Sample)) {
			printf("\nError reading input wavefile!\n\n");
			return FALSE;
		}
		L = abs((short)LOWORD(Sample)); // Left
		R = abs((short)HIWORD(Sample)); // Right

		if (L > WIMax) WIMax = L;
		if (R > WIMax) WIMax = R;

		if (!fQuite)
			if (++j == n) {
				j = 0;
				printf(" reading input for normalization... %i%%\r", ++p);
			}
	}

	if (!fQuite) {
		printf("                                             \r");
		printf("   %10.4f normalization gain factor\n", MAXSHORT/(double)WIMax);
	}
	
	// Close data chunks
	mmioAscend(hWI, &WISC, 0); 
	mmioAscend(hWI, &WIGH, 0); 

	// Close wavefile
	mmioClose(hWI, 0);

	return TRUE;
}

///////////////////////////////////////////////////
// Write output wavefile, while processing
// normalization and adjusting sample length
// given by Ration1/Ration2 or new length.
// Returns TRUE on success.
BOOL ProcessOutput(char *szInputFileName, char *szOutputFileName) {
	long Sample;					// One audio sample (both chanels)
	BOOL fConvertLength;	// Convert wave length?
	BOOL fExpand;					// Expand output file (Ration1 > Ration2)
	long i, n, j, m;
	__int64 p;
	BOOL geti, puto;

	// Open input wavefile
	if (!(hWI = mmioOpen(szInputFileName, NULL, MMIO_READ | MMIO_DENYWRITE)))	{
		printf("Error opening input wavefile '%s'.\n\n", szInputFileName);
		return FALSE;
	}

	// Set a large read buffer
	if (mmioSetBuffer(hWI, NULL, FILEIOBUFFERSIZE, 0)) {
		printf("Error allocating write buffer!\n\n");
		return FALSE;
	}

	// Locate a WAVE Group header in the file, and read it in.
	// This marks the start of any embedded WAVE format within the file.
	WIGH.fccType = FOURCC_WAVE; 
	if (mmioDescend(hWI, (LPMMCKINFO)&WIGH, NULL, MMIO_FINDRIFF)) {
		printf("Error reading input wavefile - is it a wavefile?\n\n");
		return FALSE;
	}

	// Locate "data" chunk it contains the audio information
	WISC.ckid = FOURCC_data; 
	if (mmioDescend(hWI, &WISC, &WIGH, MMIO_FINDCHUNK)) { 
		printf("Error - input wavefile has no data chunk!"); 
		return FALSE;
	}

	// Open the output file.
	if (!(hWO = mmioOpen(szOutputFileName, NULL, MMIO_CREATE | MMIO_WRITE | MMIO_EXCLUSIVE)))	{
		printf("Error creating output wavefile '%s'.\n\n", szOutputFileName);
		return FALSE;
	}
	
	// Set a large write buffer
	if (mmioSetBuffer(hWO, NULL, FILEIOBUFFERSIZE, 0)) {
		printf("Error allocating write buffer!\n\n");
		return FALSE;
	}
	// Write RIFF header
	WOGH.fccType = FOURCC_WAVE;
	if (mmioCreateChunk(hWO, &WOGH, MMIO_CREATERIFF)) {
		printf("Error - output wavefile can not be written!\n\n"); 
		return FALSE;
	}

	// Write wave format chunk
	WOSC.ckid = FOURCC_fmt; 
	if (mmioCreateChunk(hWO, &WOSC, 0)) {
		printf("Error - output wavefile can not be written!\n\n"); 
		return FALSE;
	}
	// Write wave format
	if (mmioWrite(hWO, (HPSTR)&WIFormat, sizeof(WIFormat) - sizeof(WORD)) !=
		(LRESULT)sizeof(WIFormat) - sizeof(WORD)) {
		printf("Error writing output wavefile!\n\n");
		return FALSE;
	}
	// Close format chunk
	mmioAscend(hWO, &WOSC, 0); 

	// Write wave data chunk
	WOSC.ckid = FOURCC_data; 
	WOSC.cksize = 0;
	if (mmioCreateChunk(hWO, &WOSC, 0)) {
		printf("Error - output wavefile can not be written!\n\n"); 
		return FALSE;
	}

	// Setup conditions
	if (Ration1 != 0) { // Make wave shorter or longer by given ration
		fConvertLength = TRUE;
		if (Ration2 == 0)
			Ration2 = WISamples;
		fExpand = (Ration1 > Ration2);
		if (fExpand) {
			i = Ration1;
			Ration1 = Ration2;
			Ration2 = i;
		}
		p = (__int64)2 * Ration1 - Ration2;
	}
	else {
		fConvertLength = FALSE;
		geti = TRUE; // Read all input samples
		puto = TRUE; // Write all output samples
	}

	if (WOSilence > 0) { // Add silence to beginning of output file
		if (!fQuite)
			printf(" adding %i samples silence to output wavefile...\r", WOSilence);
		Sample = 0;
		for (i = 0; i < WOSilence; i++) {
			// Write silence sample
			if (mmioWrite(hWO, (HPSTR)&Sample, sizeof(Sample)) != (LRESULT)sizeof(Sample)) {
				printf("\nError writing output wavefile!\n\n");
				return FALSE;
			}
			WOSamples++;
		}
	}

	if (!fQuite) {
		ldiv_t t;
		
		t = ldiv(WISamples, 100);
		n = t.quot;
		printf(" converting and writing output wavefile... 0%%               \r");
	}

	for (j = 0, m = 0, i = 0; i < WISamples; i++) {
		if (fConvertLength) {
			if (p < 0) { // Drop a sample in input (expand) or output (shorten)
				p += (__int64)2 * Ration1;
				if (fExpand)
					geti = FALSE;
				else
					puto = FALSE;
			}
			else {
				p += (__int64)2 * (Ration1 - Ration2);
				geti = TRUE;
				puto = TRUE;
			}
		}

		if (geti) { // Need input sample
			// Read sample
			if (mmioRead(hWI, (HPSTR)&Sample, sizeof(Sample)) != (LRESULT)sizeof(Sample)) {
				printf("\nError reading input wavefile!\n\n");
				return FALSE;
			}

			if (fNormalize) {
				long L, R;

				L = (short)MulDiv((short)LOWORD(Sample), (long)MAXSHORT, WIMax);
				R = (short)MulDiv((short)HIWORD(Sample), (long)MAXSHORT, WIMax);
				Sample = MAKELONG((WORD)L, (WORD)R);
			}
		}
		else { // Drop input sample
			i--; // Don't account for input sample
			if (!fQuite)
				j--; // Don't update meter
		}

		if (puto) { // Sample is going to output
			if (WODrop > 0)
				WODrop--; // Drop a an output sample
			else {
				// Write sample to output wavefile
				if (mmioWrite(hWO, (HPSTR)&Sample, sizeof(Sample)) != (LRESULT)sizeof(Sample)) {
					printf("\nError writing output wavefile!\n\n");
					return FALSE;
				}
				WOSamples++;
			}
		}

		if (!fQuite)
			if (++j == n) {
				j = 0;
				printf(" converting and writing output wavefile... %i%%\r", ++m);
			}
	}

	if (!fQuite) {
		printf("                                                 \r");
		printf("   %10i samples total in output wavefile\n", WOSamples);
	}
	
	// Close output chunks
	mmioAscend(hWO, &WOSC, 0); 
	mmioAscend(hWO, &WOGH, 0); 

	// Close input chunks
	mmioAscend(hWI, &WISC, 0); 
	mmioAscend(hWI, &WIGH, 0); 

	// Close wavefiles
	mmioClose(hWO, 0);
	mmioClose(hWI, 0);

	return TRUE;
}

///////////////////////////////////////////////////
// Checks the input wave for correct type.
// Fill in structure WIFormat structure.
// Returns TRUE on success.
BOOL CheckInputWave(char *szFileName) {
	// Open input wavefile
	if (!(hWI = mmioOpen(szFileName, NULL, MMIO_READ | MMIO_ALLOCBUF | MMIO_DENYWRITE)))	{
		printf("Error opening input wavefile '%s'.\n\n", szFileName);
		return FALSE;
	}

	// Locate a WAVE Group header in the file, and read it in.
	WIGH.fccType = FOURCC_WAVE; 
	if (mmioDescend(hWI, (LPMMCKINFO)&WIGH, NULL, MMIO_FINDRIFF)) {
		printf("Error reading input wavefile - is it a wavefile?\n\n");
		return FALSE;
	}

	// Locate the WAVE's "fmt " chunk and read in its header
	WISC.ckid = FOURCC_fmt; 
	if (mmioDescend(hWI, &WISC, &WIGH, MMIO_FINDCHUNK)) {
		printf("Error reading input wavefile - no format found!\n\n");
		return FALSE;
	}

	// Read in the "fmt " chunk into a WAVEFORMATEX structure
	if (mmioRead(hWI, (HPSTR)&WIFormat, WISC.cksize) != (LRESULT)WISC.cksize) {
		printf("Error reading input wavefile - format unknown!\n\n");
		return FALSE;
	}

	// Close format chunk
	mmioAscend(hWI, &WISC, 0); 

	// Check for correct format, only PCM 16bit stereo
	if (WIFormat.wFormatTag != WAVE_FORMAT_PCM) {
		printf("Error - input wavefile format must be PCM!\n\n");
		return FALSE;
	}
	if (WIFormat.nChannels != 2) {
		printf("Error - input wavefile format must be stereo!\n\n");
		return FALSE;
	}
	if (WIFormat.wBitsPerSample != 16) {
		printf("Error - input wavefile format must be 16 bit!\n\n");
		return FALSE;
	}

	// Locate "data" chunk it contains the audio information
	WISC.ckid = FOURCC_data; 
	if (mmioDescend(hWI, &WISC, &WIGH, MMIO_FINDCHUNK)) { 
		printf("Error - input wavefile has no data chunk!"); 
		return FALSE;
	}
	WISamples = WISC.cksize / WIFormat.nBlockAlign;
	if (WISamples == 0L) {
		printf("Error - input wavefile has no audio data!\n\n");
		return FALSE;
	}

	// Close data chunks
	mmioAscend(hWI, &WISC, 0); 
	mmioAscend(hWI, &WIGH, 0); 

	// Close wavefile
	mmioClose(hWI, 0);

	return TRUE;
}

///////////////////////////////////////////////////
// Check if we can create output wavefile.
// Returns TRUE on success
BOOL CheckOutputWave(char *szOutputFile) {
	// Open the output file.
	if (!(hWO = mmioOpen(szOutputFile, NULL, MMIO_CREATE | MMIO_ALLOCBUF | MMIO_WRITE | MMIO_EXCLUSIVE)))	{
		printf("Error creating output wavefile '%s'.\n\n", szOutputFile);
		return FALSE;
	}

	// Close wavefile
	mmioClose(hWO, 0);

	return TRUE;
}

///////////////////////////////////////////////////
// Print usage message
void Usage(void) {
	printf("Usage: WLA <File In> <File Out> [-r] [-l] [-n] [-q]\n");
	printf("   <File In>:  Source wavefilename.\n");
	printf("   <File Out>: Destination wavefilename.\n");
	printf("   [-r]:       Change ration: -r 2 5 keeps every 2 samples out of 5.\n");
	printf("   [-l]:       New length in samples: -l 5000 gives output of 5000 samples.\n");
	printf("   [-n]:       Nomalize amplitude.\n");
	printf("   [-s]:       Silence: -s 44100 adds 44100 samples of silence to the\n");
	printf("               beginning of destination wavefile.\n");
	printf("   [-d]:       Drop samples: -d 100 drops the first 100 samples of destination.\n");
	printf("               wavefile.\n");
	printf("   [-q]:       Quit operation.\n\n");
}

///////////////////////////////////////////////////
// Prints parameter err
void ParameterError(const char *msg) {
	Usage();
	printf("%s\n\n", msg);
}


///////////////////////////////////////////////////
// Main
int main(int argc, char* argv[])
{
	int i;

	// Copyright
	printf("Wave Length Adjust (WLA) version %s\n",Version);
	printf("Copyright (c) 2001 by Joakim Hjort. Distributed under GPL.\n\n");

	if (argc < 3) {
		Usage();
		goto ErrorExit;
	}

	if (!CheckInputWave(argv[1]))
		goto ErrorExit;

	if (!CheckOutputWave(argv[2]))
		goto ErrorExit;

	// Read and process all parameters
	for (i = 3; i < argc; i++) {
		if (argv[i][0] == '-') {
			switch (argv[i][1]) {
				case 'r':	// Ration settings
					if (argc < i + 2) {
						ParameterError("[-r] parameter error, expected to ration values.");
						goto ErrorExit;
					}
					sscanf(argv[i + 1], "%d", &Ration1);
					sscanf(argv[i + 2], "%d", &Ration2);
					if (Ration1 <= 0 || Ration2 <= 0) {
						ParameterError("[-r] parameter error, must be positive number.");
						goto ErrorExit;
					}
					i += 2;
					break;

				case 'l':	// New sample length
					if (argc < i + 1) {
						ParameterError("[-l] parameter error, expected new length in samples.");
						goto ErrorExit;
					}
					sscanf(argv[i + 1], "%d", &Ration1);
					if (Ration1 <= 0) {
						ParameterError("[-l] parameter error, value must be a hole positive number.");
						goto ErrorExit;
					}
					i++;
					break;

				case 's':	// Add silence to output wavefile
					if (argc < i + 1) {
						ParameterError("[-s] parameter error, expected silence length in samples.");
						goto ErrorExit;
					}
					sscanf(argv[i + 1], "%d", &WOSilence);
					if (WOSilence <= 0) {
						ParameterError("[-s] parameter error, value must be a positive number.");
						goto ErrorExit;
					}
					i++;
					break;

				case 'd':	// Drop first samples in output file
					if (argc < i + 1) {
						ParameterError("[-d] parameter error, expected drop length in samples.");
						goto ErrorExit;
					}
					sscanf(argv[i + 1], "%d", &WODrop);
					if (WODrop <= 0) {
						ParameterError("[-d] parameter error, value must be a positive number.");
						goto ErrorExit;
					}
					i++;
					break;

				case 'n':	// Normalize
					fNormalize = TRUE;
					break;

				case 'q':	// Quit operation
					fQuite = TRUE;
					break;

				default:
					ParameterError("Unknown parameter.");
					goto ErrorExit;
					break;
			}
		}
	}
	
	// Show format
	if (!fQuite) {
		printf("Input wavefile '%s':\n", argv[1]);
		printf("   %10s type\n", "PCM"); // WIFormat.wFormatTag == 1
		printf("   %10d channel(s)\n", WIFormat.nChannels);
		printf("   %10ld samples/sec\n", WIFormat.nSamplesPerSec);
		printf("   %10ld avarage bytes/sec\n", WIFormat.nAvgBytesPerSec);
		printf("   %10d bytes/sample\n", WIFormat.nBlockAlign);
		printf("   %10d bits/sample\n", WIFormat.wBitsPerSample);
		printf("   %10d sampels total\n", WISamples);
	}

	if (fNormalize) 
		if (!Normalize(argv[1]))
			goto ErrorExit;

	if (!ProcessOutput(argv[1], argv[2]))
			goto ErrorExit;
		
	return 0;		

ErrorExit:

	if (hWI != 0)	mmioClose(hWI, 0);
	if (hWO != 0)	mmioClose(hWO, 0);
	return -1;
}
